use crate::Error;
use futures::Future;
use futures_util::{SinkExt, StreamExt};
use nostr_rs_types::{Request, Response};
use parking_lot::RwLock;
use std::{pin::Pin, sync::Arc};
use tokio::{
    sync::{broadcast, mpsc, oneshot},
    time::{sleep, timeout, Duration},
};
use tokio_tungstenite::{connect_async, tungstenite::Message};
use url::Url;

/// Relayer object
#[derive(Debug)]
pub struct Relayer {
    /// URL of the relayer
    pub url: String,
    /// Sender to the relayer. This can be used to send a Requests to this relayer
    pub send_to_socket: mpsc::Sender<Request>,
    /// Internal receiver. This is used to receive messages from the relayer
    recv_from_socket: broadcast::Receiver<(Response, String)>,
    /// This sender signals to background connection to stop
    stop_service: oneshot::Sender<()>,
}

const NO_ACTIVITY_TIMEOUT_SECS: u64 = 30;

impl Relayer {
    /// Creates a new relayer
    pub fn new<F>(
        sent_messages: Arc<RwLock<Vec<Request>>>,
        connection_retries: u16,
        url: &str,
        on_connection: Option<F>,
    ) -> Result<Self, Error>
    where
        F: (Fn(Vec<Request>, mpsc::Sender<Request>) -> Pin<Box<dyn Future<Output = ()> + Send>>)
            + Send
            + Sync
            + 'static,
    {
        let (send_to_socket, receiver) = mpsc::channel(100_000);
        let (recv_from_socket, stop_service) = Self::spawn_background_client(
            sent_messages,
            send_to_socket.clone(),
            receiver,
            url,
            connection_retries,
            on_connection,
        )?;

        Ok(Self {
            url: url.to_owned(),
            send_to_socket,
            stop_service,
            recv_from_socket,
        })
    }

    fn spawn_background_client<F>(
        sent_messages: Arc<RwLock<Vec<Request>>>,
        send_to_socket: mpsc::Sender<Request>,
        mut receiver: mpsc::Receiver<Request>,
        url_str: &str,
        connection_retries: u16,
        on_connection: Option<F>,
    ) -> Result<(broadcast::Receiver<(Response, String)>, oneshot::Sender<()>), Error>
    where
        F: (Fn(Vec<Request>, mpsc::Sender<Request>) -> Pin<Box<dyn Future<Output = ()> + Send>>)
            + Send
            + Sync
            + 'static,
    {
        let (publish_to_listener, recv_from_socket) = broadcast::channel(10_000);
        let (stop_service, mut stopper_recv) = oneshot::channel();

        let url = url_str.to_owned();
        let url_parsed = Url::parse(&url)?;

        tokio::spawn(async move {
            let mut reconnect = true;
            let mut retries = 0;

            while reconnect && retries <= connection_retries {
                log::warn!("{}: Connect attempt {}", url, retries);
                retries += 1;
                let mut socket = if let Ok(x) = connect_async(url_parsed.clone()).await {
                    x.0
                } else {
                    log::warn!("{}: Failed to connect", url);
                    sleep(Duration::from_secs(5)).await;
                    continue;
                };

                log::info!("Connected to {}", url);

                if let Some(on_connection) = &on_connection {
                    let sent_messages = sent_messages.read().to_owned();
                    on_connection(sent_messages, send_to_socket.clone()).await;
                }

                loop {
                    tokio::select! {
                        Ok(()) = &mut stopper_recv => {
                            log::warn!("{}: Breaking client due external signal", url);
                            reconnect = false;
                            break;
                        },

                        Some(msg) = receiver.recv() => {
                            if let Ok(json) = serde_json::to_string(&msg) {
                                log::info!("{}: Sending {}", url, json);
                                if let Err(x) = socket.send(Message::Text(json)).await {
                                    log::error!("{} : Reconnecting due {}", url, x);
                                    break;
                                }
                            }
                        }
                        msg = timeout(Duration::from_secs(NO_ACTIVITY_TIMEOUT_SECS), socket.next()) => {
                            let msg = if let Ok(Some(Ok(msg))) = msg {
                                    if let Ok(msg) = msg.into_text() {
                                        msg
                                    } else {
                                        continue;
                                    }
                                } else {
                                    log::error!("{} Reconnecting client due of empty recv: {:?}", url, msg);
                                    break;
                                };

                            if msg.is_empty() {
                                continue;
                            }

                            log::info!("New message: {}", msg);
                            retries = 0;


                            let msg: Result<Response, _> = serde_json::from_str(&msg);

                            if let Ok(msg) = msg {
                                if let Err(error) = publish_to_listener.send((msg, url.to_owned())) {
                                    log::error!("{}: Reconnecting client because of {}", url, error);
                                    break;
                                }
                            }
                        }
                        else => {
                            log::warn!("{}: else", url);
                            break;
                        }
                    }
                    // Throttle down to not spam the server with reconnections
                    sleep(Duration::from_millis(500)).await;
                }
            }

            log::warn!("{}: Disconnected", url);
        });

        Ok((recv_from_socket, stop_service))
    }

    /// Checks if the relayer background connection is running
    pub fn is_running(&self) -> bool {
        !self.stop_service.is_closed()
    }

    /// Subscribe to responses sent through this relayer. If the relayer is not
    /// running, this will return None
    pub fn subscribe(&self) -> Option<broadcast::Receiver<(Response, String)>> {
        if self.stop_service.is_closed() {
            None
        } else {
            Some(self.recv_from_socket.resubscribe())
        }
    }

    /// Sends a requests to this relayer
    pub async fn send(&self, request: Request) -> Result<(), Error> {
        Ok(self.send_to_socket.send(request).await?)
    }

    /// Stops the background thread that has the connection to this relayer
    pub async fn stop(self) {
        let _ = self.stop_service.send(());
    }
}
