//! Relayers
//!
//! This is the main entry point to the client library.
use crate::{Error, Relayer};
use futures::Future;
use futures_util::{stream::FuturesUnordered, StreamExt};
use nostr_rs_types::{Request, Response};
use parking_lot::RwLock;
use std::{collections::HashMap, pin::Pin, sync::Arc};
use tokio::sync::mpsc;

/// Clients
///
/// This is a set of outgoing connections to relayers. This struct can connect
/// async to N relayers offering a simple API to talk to all of them at the same
/// time, and to receive messages
#[derive(Debug, Clone)]
pub struct Relayers {
    clients: Arc<RwLock<HashMap<String, Relayer>>>,
    messages: Arc<RwLock<Vec<Request>>>,
}

impl Default for Relayers {
    fn default() -> Self {
        Self {
            clients: Arc::new(RwLock::new(HashMap::new())),
            messages: Arc::new(RwLock::new(vec![])),
        }
    }
}

impl Relayers {
    /// Receives a Response from any of the connected relayers
    pub async fn recv(&self) -> Result<Option<(Response, String)>, Error> {
        let mut subscriptions = self
            .clients
            .read()
            .iter()
            .map(|(_, c)| c.subscribe())
            .collect::<Vec<Option<_>>>()
            .into_iter()
            .filter_map(|x| x)
            .collect::<Vec<_>>();

        if subscriptions.is_empty() {
            return Err(Error::Disconnected);
        }

        let mut futures = FuturesUnordered::new();

        for sub in subscriptions.iter_mut() {
            futures.push(sub.recv());
        }

        if let Some(Ok(response)) = futures.next().await {
            Ok(Some(response))
        } else {
            Ok(None)
        }
    }

    /// Attempts to receive a Response from any of the relayers but won't wait if none
    /// is available. It will return None if None is available
    pub fn try_recv(&self) -> Result<Option<(Response, String)>, Error> {
        let mut subscriptions = self
            .clients
            .read()
            .iter()
            .map(|(_, c)| c.subscribe())
            .collect::<Vec<Option<_>>>()
            .into_iter()
            .filter_map(|x| x)
            .collect::<Vec<_>>();

        if subscriptions.is_empty() {
            return Err(Error::Disconnected);
        }

        for sub in subscriptions.iter_mut() {
            if let Ok(msg) = sub.try_recv() {
                return Ok(Some(msg));
            }
        }
        Ok(None)
    }

    /// Returns the number of active connections. If a connection to a relayer
    /// is not active it will be removed from the list
    pub fn check_active_connections(&self) -> usize {
        let mut clients = self.clients.write();
        let mut to_remove = vec![];
        for (url, client) in clients.iter() {
            if !client.is_running() {
                to_remove.push(url.to_owned());
            }
        }

        for url in to_remove.iter() {
            clients.remove(url);
        }

        clients.len()
    }

    /// Sends a request to all the connected relayers
    pub async fn send(&self, request: Request) {
        let senders = self
            .clients
            .read()
            .iter()
            .map(|(_, c)| c.send_to_socket.clone())
            .collect::<Vec<mpsc::Sender<_>>>();

        // Add a copy of the request for incoming connections.
        self.messages.write().push(request.clone());

        for sender in senders.iter() {
            let _ = sender.send(request.clone()).await;
        }
    }

    /// Creates a connection to a new relayer. If the connection is successful a
    /// Callback will be called, with a list of previously sent requests, and a
    /// Sender to send new requests to this relayer alone.
    ///
    /// The same callback will be called for every reconnection to the same
    /// relayer, also the callback will be called, giving the chance to re-send
    /// sent requests to the new connections
    ///
    /// This function will open a connection at most once, if a connection
    /// already exists false will be returned
    pub async fn connect_to<F>(
        &self,
        url: &str,
        connection_retries: u16,
        on_connection: Option<F>,
    ) -> Result<bool, Error>
    where
        F: (Fn(Vec<Request>, mpsc::Sender<Request>) -> Pin<Box<dyn Future<Output = ()> + Send>>)
            + Send
            + Sync
            + 'static,
    {
        let mut clients = self.clients.write();
        Ok(if clients.get(url).is_some() {
            false
        } else {
            log::warn!("Connecting to {}", url);
            clients.insert(
                url.to_owned(),
                Relayer::new(
                    self.messages.clone(),
                    connection_retries,
                    url,
                    on_connection,
                )?,
            );
            true
        })
    }

    /// Disconnects from all relayers
    pub async fn stop(&self) {
        let mut clients = self.clients.write();
        let keys = clients.keys().cloned().collect::<Vec<_>>();

        for key in keys.iter() {
            if let Some(client) = clients.remove(key) {
                client.stop().await;
            }
        }
    }
}
